from quiffen import utils
from quiffen.core.accounts import Account
from quiffen.core.categories_classes import Category, Class
from quiffen.core.transactions import Transaction, Investment

try:
    import pandas as pd

    PANDAS_INSTALLED = True
except ModuleNotFoundError:
    PANDAS_INSTALLED = False

VALID_TRANSACTION_ACCOUNT_TYPES = [
    '!Type:Cash',
    '!Type:Bank',
    '!Type:Ccard',
    '!Type:Oth A',
    '!Type:Oth L',
    '!Type:Invoice'
]


class ParserException(Exception):
    pass


class Qif:
    """
    The main class of the package. For parsing QIF files.

    See the readme for usage examples.

    """

    def __init__(self,
                 accounts: dict = None,
                 categories: dict = None,
                 classes: dict = None
                 ):
        """Initialise an instance of the Qif class.

        Parameters
        ----------
        accounts : dict, default=None
            A dict of accounts in the form {'Account Name': account_object}.
        categories : dict, default=None
            A dict of categories in the form {'Category Name': category_object}.
        classes : dict, default=None
            A dict of classes in the form {'Class Name': class_object}.

        Raises
        ------
        TypeError
            If any provided arguments are the wrong type.
        """
        if accounts:
            self._assert_type(accounts, Account)
        else:
            accounts = {}

        if categories:
            self._assert_type(categories, Category)
        else:
            categories = {}

        if classes:
            self._assert_type(classes, Class)
        else:
            classes = {}

        self._accounts = accounts
        self._categories = categories
        self._classes = classes

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __str__(self):
        accounts = {name: repr(account) for (name, account) in self._accounts.items()}
        categories = {name: repr(cat) for (name, cat) in self._categories.items()}
        return f"""
QIF:
    Accounts: {accounts}
    Categories: {categories}
    Classes: {[repr(klass) for klass in self._classes]}
"""

    def __repr__(self):
        accounts = {name: repr(account) for (name, account) in self._accounts.items()}
        categories = {name: repr(cat) for (name, cat) in self._categories.items()}
        return f'Qif(accounts={accounts}, ' \
               f'categories={categories}, ' \
               f'classes={[repr(klass) for klass in self._classes]})'

    @property
    def accounts(self):
        return self._accounts

    @accounts.setter
    def accounts(self, new_accounts):
        self._assert_type(new_accounts, Account)
        self._accounts = new_accounts

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, new_categories):
        self._assert_type(new_categories, Category)
        self._categories = new_categories

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, new_classes):
        self._assert_type(new_classes, Class)
        self._classes = new_classes

    @classmethod
    def parse(cls, filepath, separator='\n', day_first=True):
        """Return a class instance from a QIF file.

        Parameters
        ----------
        filepath : str
            The path to the QIF file.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.
        day_first : bool, default=True
             Whether the day or month comes first in the date.

        Returns
        -------
        Qif
            A Qif object containing all the data in the QIF file.
        """
        data = cls._read_qif(filepath)

        accounts = {}
        last_account = None
        categories = {}
        classes = {}

        sections = data.split('^\n')
        last_header = None
        line_number = 1

        for section in sections:
            if not section:
                continue

            header_line = section.split('\n')[0]
            if header_line is None:
                continue

            # Allow for comments and blank lines at the top of sections
            i = 0
            while True:
                if header_line.strip() != '':
                    if header_line[0] != '#':
                        break
                i += 1
                print(i, section.split('\n'))
                header_line = section.split('\n')[i]

            # Check for new categories and accounts first, as then we can be sure it's a transaction in case a default
            # account needs to be added
            if header_line[0] != '!':
                if last_header is None:
                    raise ParserException(f'Header \'{header_line}\' not recognised and no previous header supplied')
                header_line = last_header

            if '!Type:Cat' in header_line:
                # Section contains category information
                new_category = Category.from_string(section)
                categories = utils.create_categories(new_category, categories)
            elif '!Type:Class' in header_line:
                new_class = Class.from_string(section)
                classes[new_class.name] = new_class
            elif '!Account' in header_line:
                new_account = Account.from_string(section)
                accounts[new_account.name] = new_account
                last_account = new_account.name
            elif '!Type' in header_line and not accounts:
                # Accounts is empty and there's a transaction, so create default account to put transactions in
                default_account = Account(name='Quiffen Default Account',
                                          desc='The default account created by Quiffen when no other accounts were '
                                               'present')
                accounts[default_account.name] = default_account
                last_account = default_account.name
            elif '!Type:Invst' in header_line:
                # Investment
                new_investment = Investment.from_string(section, separator=separator, day_first=day_first,
                                                        line_number=line_number)
                accounts[last_account].add_transaction(new_investment, 'Invst')
            elif header_line in VALID_TRANSACTION_ACCOUNT_TYPES:
                # Other transaction type
                new_transaction, new_categories, new_classes = Transaction.from_string(section, separator=separator,
                                                                                       day_first=day_first,
                                                                                       line_number=line_number)
                accounts[last_account].add_transaction(new_transaction, header_line)
                categories.update(new_categories)
                classes.update(new_classes)

            last_header = header_line
            line_number += len(section.split('\n'))

        return cls(accounts=accounts, categories=categories, classes=classes)

    @staticmethod
    def _assert_type(iterable, types):
        # Assert that all items in an iterable are of specific types.
        if isinstance(iterable, dict):
            for item in iterable.values():
                if not isinstance(item, types):
                    raise TypeError(f'\'{repr(item)} is not of type {types}')
        else:
            for item in iterable:
                if not isinstance(item, types):
                    raise TypeError(f'\'{repr(item)} is not of type {types}')

    @staticmethod
    def _read_qif(path):
        # Validate QIF file provided and return data.
        if path[-3:].lower() != 'qif':
            raise FileNotFoundError(f'\'{path}\' does not point to a valid QIF file. Only .QIF file types are allowed')

        with open(path, 'r') as f:
            data = f.read().strip().strip('\n')
            if not data:
                raise ParserException('File is empty')

        return data

    def add_account(self, new_account):
        """Add a new account to the Qif object.

        Parameters
        ----------
        new_account : Account
            The Account to be added to the Qif.

        Raises
        ------
        TypeError
            If ``new_account`` is not an Account object.
        """
        self._assert_type([new_account], Account)
        self._accounts[new_account.name] = new_account

    def remove_account(self, account_name):
        """Remove an account from this Qif object.

        Parameters
        ----------
        account_name : str
            The name of the account.

        Returns
        -------
        Account
            The Account removed.
        """
        return self._accounts.pop(account_name)

    def add_category(self, new_category):
        """Add a new category to the object.

        Parameters
        ----------
        new_category : Category
            The Category to be added to the Qif.

        Raises
        ------
        TypeError
            If ``new_category`` is not a Category object.
        """
        self._assert_type([new_category], Category)
        self._categories[new_category.name] = new_category

    def remove_category(self, category_name, keep_children=True):
        """Remove a category from the Qif object.

        Parameters
        ----------
        category_name : str
            The name of the category to be removed.
        keep_children : bool, default=True
            Whether or not the children of the category will be kept (grandparent node will become the parent node).

        Returns
        -------
        Category
            The Category removed.
        """
        if category_name in self._categories:
            if keep_children:
                self._categories.update({category.name: category for category in
                                         self._categories[category_name].children})
            return self._categories.pop(category_name)

        categories_to_visit = list(self._categories.values())

        for category in categories_to_visit:
            if category.children:
                categories_to_visit.extend(category.children)
            if category.name == category_name:

                # Change children hierarchies
                children_to_visit = category.children
                for child in children_to_visit:
                    if child.children:
                        children_to_visit.extend(child)
                    if child.hierarchy:
                        child.hierarchy = child.hierarchy.replace(f'{category.name}:', '')
                    child.parent = category.parent

                parent = category.parent
                if parent is not None:
                    category_idx = parent.children.index(category)
                    parent.children.pop(category_idx)
                    parent.children.extend(category.children)
                    category.parent = None
                return category

    def add_class(self, new_class):
        """Add a new class to the object.

        Parameters
        ----------
        new_class : Class
            The Class to be added to the Qif.

        Raises
        ------
        TypeError
            If ``new_class`` is not a Class object.
        """
        self._assert_type([new_class], Class)
        self._classes[new_class.name] = new_class

    def remove_class(self, class_name):
        """Remove a class from the Qif object.

        Parameters
        ----------
        class_name : str
            The name of the Class to be removed.

        Returns
        -------
        Class
            The Class removed.
        """
        return self._classes.pop(class_name)

    def to_qif(self, path=None, date_format='%d/%m/%Y'):
        """Write the Qif object to a QIF file and return the string.

        Parameters
        ----------
        path : str, default=None
            The path of the QIF file to be generated, if desired.
        date_format : str, default='%d/%m/%Y'
            The format of the date (using ``datetime`` verbs) to be input into the QIF file.

        Returns
        -------
        qif_data : str
            The string of QIF data.
        """
        qif_data = ''

        if self._classes:
            qif_data += '!Type:Class\n'
            for klass in self._classes.values():
                qif_data += f'N{klass.name}\n'
                if klass.desc:
                    qif_data += f'D{klass.desc}\n'
                qif_data += '^\n'

        if self._categories:
            categories_to_visit = list(self._categories.values())
            qif_data += '!Type:Cat\n'
            for category in categories_to_visit:
                if category.children:
                    categories_to_visit.extend(category.children)

                qif_data += f'N{category.hierarchy}\n'

                if category.desc:
                    qif_data += f'D{category.desc}\n'

                if category.tax_related is not None:
                    qif_data += f'T{category.tax_related}\n'

                if category.expense is not None:
                    qif_data += f'E{category.expense}\n'

                if category.income is not None:
                    qif_data += f'I{category.income}\n'

                if category.budget_amount is not None:
                    qif_data += f'B{category.budget_amount}\n'

                if category.tax_schedule_info:
                    qif_data += f'R{category.tax_schedule_info}\n'

                qif_data += '^\n'

        if self._accounts:
            for account in self._accounts.values():
                qif_data += '!Account\n'

                qif_data += f'N{account.name}\n'

                if account.desc:
                    qif_data += f'D{account.desc}\n'

                if account.account_type:
                    qif_data += f'T{account.account_type}\n'

                if account.credit_limit is not None:
                    qif_data += f'L{account.credit_limit}\n'

                if account.balance is not None:
                    qif_data += f'${account.balance}\n'

                if account.date_at_balance:
                    qif_data += f'/{account.date_at_balance.strftime(date_format)}\n'

                qif_data += '^\n'

                for (header, transaction_list) in account.transactions.items():
                    qif_data += f'!Type:{header}\n'

                    for transaction in transaction_list:
                        qif_data += f'D{transaction.date.strftime(date_format)}\n'

                        if transaction.amount is not None:
                            qif_data += f'T{transaction.amount}\n'

                        if transaction.memo:
                            qif_data += f'M{transaction.memo}\n'

                        if transaction.cleared:
                            qif_data += f'C{transaction.cleared}\n'

                        if 'Invst' not in header:
                            if transaction.payee:
                                qif_data += f'P{transaction.payee}\n'

                            if transaction.payee_address:
                                qif_data += f'A{transaction.payee_address}\n'

                            if transaction.category:
                                qif_data += f'L{transaction.category.hierarchy}\n'

                            if transaction.to_account:
                                qif_data += f'L[{transaction.to_account}]\n'

                            if transaction.check_number is not None:
                                qif_data += f'N{transaction.check_number}\n'

                            if transaction.reimbursable_expense is not None:
                                qif_data += f'F{transaction.reimbursable_expense}\n'

                            if transaction.first_payment_date:
                                qif_data += f'1{transaction.first_payment_date.strftime(date_format)}\n'

                            if transaction.loan_length is not None:
                                qif_data += f'2{transaction.loan_length}\n'

                            if transaction.num_payments is not None:
                                qif_data += f'3{transaction.num_payments}\n'

                            if transaction.periods_per_annum is not None:
                                qif_data += f'4{transaction.periods_per_annum}\n'

                            if transaction.interest_rate is not None:
                                qif_data += f'5{transaction.interest_rate}\n'

                            if transaction.current_loan_balance is not None:
                                qif_data += f'6{transaction.current_loan_balance}\n'

                            if transaction.original_loan_amount is not None:
                                qif_data += f'7{transaction.original_loan_amount}\n'

                            if transaction.is_split:
                                for split in transaction.splits:
                                    if split.category:
                                        qif_data += f'S{split.category.hierarchy}\n'

                                    if split.date:
                                        qif_data += f'D{split.date.strftime(date_format)}\n'

                                    if split.amount is not None:
                                        qif_data += f'${split.amount}\n'

                                    if split.percent is not None:
                                        qif_data += f'%{split.percent}\n'

                                    if split.memo:
                                        qif_data += f'E{split.memo}\n'

                                    if split.cleared:
                                        qif_data += f'C{split.cleared}\n'

                                    if split.payee_address:
                                        qif_data += f'A{split.payee_address}\n'

                                    if split.to_account:
                                        qif_data += f'L[{split.to_account}]\n'

                                    if split.check_number is not None:
                                        qif_data += f'N{split.check_number}\n'

                        else:
                            if transaction.action:
                                qif_data += f'N{transaction.action}\n'

                            if transaction.security:
                                qif_data += f'Y{transaction.security}\n'

                            if transaction.price is not None:
                                qif_data += f'I{transaction.price}\n'

                            if transaction.quantity is not None:
                                qif_data += f'Q{transaction.quantity}\n'

                            if transaction.first_line:
                                qif_data += f'P{transaction.first_line}\n'

                            if transaction.to_account:
                                qif_data += f'L{transaction.to_account}\n'

                            if transaction.transfer_amount is not None:
                                qif_data += f'${transaction.transfer_amount}\n'

                            if transaction.commission is not None:
                                qif_data += f'O{transaction.commission}\n'

                        qif_data += '^\n'

        if path is not None:
            with open(path, 'w') as f:
                f.write(qif_data)

        return qif_data

    def to_dicts(self, data='transactions', ignore=None):
        """Return a list of dict representations of desired data.

        Parameters
        ----------
        data : {'transactions', 'investments', 'splits', 'accounts', 'categories', 'classes'}
            The data type to be converted to dicts.
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.

        Returns
        -------
        list of dict
            A list of dict objects containing ``data`` specified.

        Raises
        ------
        RuntimeError
            If ``data`` is not a valid option.
        """
        if ignore is None:
            ignore = []

        data = data.lower()
        options = ['transactions', 'investments', 'splits', 'accounts', 'categories', 'classes']

        if data not in options:
            raise RuntimeError(f'Can\'t get data for {data}. Valid options are:\n{", ".join(options)}')

        if data == 'transactions':
            transactions = []
            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for transaction in transaction_list:
                        if isinstance(transaction, Transaction):
                            transactions.append(transaction.to_dict(ignore=ignore))
            return transactions
        elif data == 'investments':
            investments = []
            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for transaction in transaction_list:
                        if isinstance(transaction, Investment):
                            investments.append(transaction.to_dict(ignore=ignore))
            return investments
        elif data == 'splits':
            splits = []
            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for transaction in transaction_list:
                        for split in transaction.splits:
                            splits.append(split.to_dict(ignore=ignore))
            return splits
        elif data == 'accounts':
            accounts = []
            for account in self._accounts.values():
                accounts.append(account.to_dict(ignore=ignore))
            return accounts
        elif data == 'categories':
            categories = []
            categories_to_visit = list(self._categories.values())

            for category in categories_to_visit:
                if category.children is not None:
                    categories_to_visit.extend(category.children)
                categories.append(category.to_dict(ignore=ignore))
            return categories
        elif data == 'classes':
            classes = []
            for klass in self._classes:
                classes.append(klass.to_dict(ignore=ignore))
            return classes

    def to_csv(self, path=None, data='transactions', ignore=None, separator=',', sub_separator=';',
               date_format='%d/%m/%Y'):
        """Write a CSV file containing desired data and return the string.

        Parameters
        ----------
        path : str, default=None
            The path of the CSV file to be created, if desired.
        data : {'transactions', 'investments', 'splits', 'accounts', 'categories', 'classes'}
            The data type to be input into the CSV file.
        ignore : list of str, default=None
            A list of strings of parameters that should be excluded from the dict.
        separator : str, default=','
            The separator for the CSV file.
        sub_separator : str, default=';'
            The character to be used to replace ``separator`` if it appears in any strings.
        date_format : str, default='%d/%m/%Y'
            The format of the date (using ``datetime`` verbs) to be input into the QIF file.

        Returns
        -------
        csv_data : str
            A string containing the CSV data.

        Raises
        ------
        RuntimeError
            If ``data`` is not a valid option.
        ValueError
            If ``separator`` and ``sub_separator`` are present within each other.
        """
        if ignore is None:
            ignore = []

        data = data.lower()
        options = ['transactions', 'splits', 'investments', 'accounts', 'categories', 'classes']

        if data not in options:
            raise RuntimeError(f'Can\'t get data for {data}. Valid options are:\n{", ".join(options)}')

        if separator in sub_separator or sub_separator in separator:
            raise ValueError('Separator and sub-separator cannot be equal or contain each other')

        csv_data = ''

        if data == 'transactions':
            headers = ['date', 'amount', 'memo', 'cleared', 'payee', 'payee_address', 'category', 'check_number',
                       'reimbursable_expense', 'small_business_expense', 'to_account', 'first_payment_date',
                       'loan_length', 'num_payments', 'periods_per_annum', 'interest_rate', 'current_loan_balance',
                       'original_loan_amount', 'is_split', 'splits']

            headers = [header for header in headers if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for transaction in transaction_list:
                        if not isinstance(transaction, Transaction):
                            continue

                        this_line = ''
                        transaction_dict = transaction.to_dict()
                        for header in headers:
                            transaction_data = transaction_dict.get(header)

                            if transaction_data is not None:
                                if 'date' in header:
                                    this_line += transaction_data.strftime(date_format)
                                elif header == 'category':
                                    this_line += transaction_data['name']
                                elif header == 'splits':
                                    this_line += str(len(transaction_data))
                                else:
                                    this_line += str(transaction_data).replace(separator, sub_separator)

                            this_line += separator
                        this_line = this_line.strip(separator) + '\n'
                        csv_data += this_line
        elif data == 'splits':
            headers = ['date', 'amount', 'memo', 'cleared', 'payee_address', 'category', 'to_account', 'check_number',
                       'percent']

            headers = [header for header in headers if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for transaction in transaction_list:
                        for split in transaction.splits:
                            this_line = ''
                            split_dict = split.to_dict()

                            for header in headers:
                                split_data = split_dict.get(header)

                                if split_data is not None:
                                    if 'date' in header:
                                        this_line += split_data.strftime(date_format)
                                    elif header == 'category':
                                        this_line += split_data['name']
                                    else:
                                        this_line += str(split_data).replace(separator, sub_separator)

                                this_line += separator
                            this_line = this_line.strip(separator) + '\n'
                            csv_data += this_line
        elif data == 'investments':
            headers = ['date', 'action', 'security', 'price', 'quantity', 'cleared', 'amount', 'memo', 'first_line',
                       'to_account', 'transfer_amount', 'commission']

            headers = [header for header in headers if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            for account in self._accounts.values():
                for transaction_list in account.transactions.values():
                    for investment in transaction_list:
                        if not isinstance(investment, Investment):
                            continue

                        this_line = ''
                        investment_dict = investment.to_dict()
                        for header in headers:
                            investment_data = investment_dict.get(header)

                            if investment_data is not None:
                                if 'date' in header:
                                    this_line += investment_data.strftime(date_format)
                                elif header == 'category':
                                    this_line += investment_data['name']
                                elif header == 'splits':
                                    this_line += str(len(investment_data))
                                else:
                                    this_line += str(investment_data).replace(separator, sub_separator)

                            this_line += separator
                        this_line = this_line.strip(separator) + '\n'
                        csv_data += this_line
        elif data == 'accounts':
            headers = ['name', 'desc', 'account_type', 'credit_limit', 'balance', 'date_at_balance', 'transactions']

            headers = [header for header in headers if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            for account in self._accounts.values():
                this_line = ''
                account_dict = account.to_dict()
                for header in headers:
                    account_data = account_dict.get(header)

                    if account_data is not None:
                        if header == 'date_at_balance':
                            this_line += account_data.strftime(date_format)
                        elif header == 'transactions':
                            this_line += str(len(account_data))
                        else:
                            this_line += str(account_data).replace(separator, sub_separator)

                    this_line += separator
                this_line = this_line.strip(separator) + '\n'
                csv_data += this_line
        elif data == 'categories':
            headers = ['name', 'desc', 'tax_related', 'expense', 'income', 'budget_amount', 'tax_schedule_info',
                       'parent', 'children']

            headers = [header for header in headers if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            categories_to_visit = list(self._categories.values())
            for category in categories_to_visit:
                if category.children is not None:
                    categories_to_visit.extend(category.children)

                this_line = ''
                category_dict = category.to_dict()
                for header in headers:
                    category_data = category_dict.get(header)

                    if category_data is not None:
                        if header == 'children':
                            this_line += str(len(category_data))
                        else:
                            this_line += str(category_data).replace(separator, sub_separator)

                    this_line += separator
                this_line = this_line.strip(separator) + '\n'
                csv_data += this_line
        elif data == 'classes':
            headers = [header for header in ['name', 'desc'] if header not in ignore]

            csv_data += separator.join(headers) + '\n'

            for klass in self._classes:
                this_line = ''
                class_dict = klass.to_dict()
                for header in headers:
                    class_data = class_dict.get(header)

                    if class_data is not None:
                        this_line += str(class_data).replace(separator, sub_separator)

                    this_line += separator
                this_line = this_line.strip(separator) + '\n'
                csv_data += this_line

        if path is not None:
            with open(path, 'w') as f:
                f.write(csv_data)

        return csv_data

    def to_dataframe(self, data='transactions', ignore=None):
        """Return a pandas DataFrame containing desired data.

        Parameters
        ----------
        data : {'transactions', 'investments', 'splits', 'accounts', 'categories', 'classes'}
            The data type to be input into the CSV file.
        ignore : list of str, default=None
            A list of strings of parameters that should be excluded from the dict.

        Returns
        -------
        DataFrame
            A DataFrame containing he desired data.

        Raises
        ------
        ModuleNotFoundError
            If the pandas module is not installed.
        """
        if ignore is None:
            ignore = []

        if not PANDAS_INSTALLED:
            raise ModuleNotFoundError('The pandas module must be installed to use this method')

        return pd.DataFrame(self.to_dicts(data=data, ignore=ignore))



from datetime import datetime
from unittest import TestCase

from quiffen.core.accounts import Account
from quiffen.core.categories_classes import Category, Class
from quiffen.core.qif import Qif
from quiffen.core.transactions import Transaction


class TestQif(TestCase):

    def test_read_qif(self):
        data = Qif._read_qif('../test.qif')

        with open('../test.qif', 'r') as f:
            data2 = f.read()

        self.assertEqual(data, data2)

    def test_add_remove_account(self):
        qif = Qif()
        acc = Account(name='Test Account')
        qif.add_account(acc)

        self.assertEqual(qif, Qif(accounts={'Test Account': Account(name='Test Account')}))

        qif.remove_account('Test Account')

        self.assertEqual(qif, Qif())

    def test_add_remove_category(self):
        qif = Qif()
        cat = Category(name='Test')
        cat2 = Category(name='Test2', hierarchy='Test:Test2')
        cat.add_child(cat2)

        qif.add_category(cat)

        self.assertEqual(qif, Qif(categories={'Test': Category(name='Test')}))

        qif.remove_category('Test')
        self.assertEqual(qif, Qif(categories={'Test2': Category(name='Test2')}))

        qif.remove_category('Test2')
        self.assertEqual(qif, Qif())

        cat.add_child(cat2)
        qif.add_category(cat)
        qif.remove_category(cat.name, keep_children=False)
        self.assertEqual(qif, Qif())

    def test_add_remove_class(self):
        qif = Qif()
        klass = Class(name='Test')

        qif.add_class(klass)
        self.assertEqual(qif, Qif(classes={'Test': Class(name='Test')}))

        qif.remove_class('Test')
        self.assertEqual(qif, Qif())

    def test_from_to_qif(self):
        qif = Qif.parse('../test.qif')
        qif.to_qif('../test2.qif')

        qif2 = Qif.parse('../test2.qif')
        self.assertEqual(qif, qif2)

    def test_to_dicts(self):
        qif = Qif()
        acc = Account(name='Test Account')
        qif.add_account(acc)
        date_now = datetime.now()

        tr = Transaction(date=date_now, amount=150)
        acc.last_header = 'Bank'
        acc.add_transaction(tr)

        tr_dicts = qif.to_dicts(data='transactions')
        self.assertEqual(tr_dicts, [{'date': date_now, 'amount': 150}])

        acc_dicts = qif.to_dicts(data='accounts')
        self.assertEqual(acc_dicts, [
            {'name': 'Test Account', 'transactions': {'Bank': [{'date': date_now, 'amount': 150}]},
             'last_header': 'Bank'}])


from setuptools import setup, find_packages

VERSION = '1.1.1'
DESCRIPTION = 'Quiffen'
with open('README.rst', 'r') as f:
    LONG_DESCRIPTION = f.read()

# Setting up
setup(
    name="quiffen",
    version=VERSION,
    author="Isaac Harris-Holt",
    author_email="isaac@harris-holt.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    install_requires=[],
    keywords=['qif', 'finance', 'data processing'],
    license='GNU GPLv3',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ],
    python_requires='>=3'
)



from abc import ABC
from datetime import datetime
import collections

from decimal import Decimal

from quiffen.utils import parse_date, create_categories
from quiffen.core.categories_classes import Category, Class


class TransactionList(collections.MutableSequence, ABC):
    """
    A class to store Transaction-type objects only in an ordered list.

    Extends collections.MutableSequence, so all list operations will work on this class.

    Attributes
    ----------
    list : list
        The list of Transaction-type objects stored by the object.
    """

    def __init__(self, *args):
        self._list = []
        self.extend(list(args))

    def __len__(self):
        return len(self._list)

    def __getitem__(self, key):
        return self._list[key]

    def __delitem__(self, key):
        del self._list[key]

    def __setitem__(self, key, value):
        self._check_if_transaction(value)
        self._list[key] = value

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return f'TransactionList({", ".join([repr(item) for item in self._list])})'

    def insert(self, key, value):
        """Insert an object into the list at a specified key. Overrides the MutableSequence.insert() method.

        Parameters
        ----------
        key : int
            The key at which to insert ``value``.
        value : Transaction or Investment
            The Transaction-type object to be inserted into the list.
        """
        self._check_if_transaction(value)
        self.list.insert(key, value)

    @staticmethod
    def _check_if_transaction(item):
        # Check if a given item is an instance of the Transaction or Investment classes and raise an error if not.
        if not isinstance(item, (Transaction, Investment, Split)):
            raise TypeError(f'Only Transaction-type objects can be appended to a TransactionList')

    @property
    def list(self):
        return self._list

    @list.setter
    def list(self, new_list):
        for item in new_list:
            self._check_if_transaction(item)
        self._list = new_list


class Transaction:
    """
    A class used to represent a transaction.

    Attributes
    ----------
    is_split : bool
        Whether or not this transaction is split (see documentation on parameter ``splits``)
    split_categories : dict
        The categories of the splits for this transaction, if transaction ``is_split``.

    Examples
    --------
    Create a Transaction instance, then convert to a dict, ignoring the date.

    >>> import quiffen
    >>> import decimal
    >>> from datetime import datetime
    >>> cat = quiffen.Category('Finances')
    >>> tr = quiffen.Transaction(date=datetime.now(), amount=decimal.Decimal(150.60), category=cat)
    >>> tr
    Transaction(date=datetime.datetime(2021, 7, 5, 10, 45, 40, 48195), amount=150.6, category=Category(name='Finances',
    expense=True, hierarchy='Finances'))
    >>> print(tr)
    Transaction:
        Date: 2021-07-05 10:45:40.048195
        Amount: 150.6
        Category: Finances
    >>> tr.to_dict(ignore=['date'], dictify_category=True)
    {'amount': 150.6, 'category': {'name': 'Finances', 'expense': True, 'income': False, 'hierarchy': 'Finances',
    'children': []}}
    """

    def __init__(self,
                 date: datetime,
                 amount: Decimal,
                 memo: str = None,
                 cleared: str = None,
                 payee: str = None,
                 payee_address: str = None,
                 category: Category = None,
                 check_number: int = None,
                 reimbursable_expense: bool = None,
                 small_business_expense: bool = None,
                 to_account: str = None,
                 first_payment_date: datetime = None,
                 loan_length: Decimal = None,
                 num_payments: int = None,
                 periods_per_annum: int = None,
                 interest_rate: float = None,
                 current_loan_balance: float = None,
                 original_loan_amount: float = None,
                 line_number: int = None,
                 splits: TransactionList = None
                 ):
        """Initialise an instance of the Transaction class.

        Parameters
        ----------
        date : datetime
            Date transaction occurred. May or may not include timestamp.
        amount : decimal.Decimal
            The amount of the transaction. May be positive or negative.
        memo : str, default=None
            Also known as the reference. A string describing the purpose behind the transaction.
        cleared : str, default=None
            The cleared status of this transaction. See the QIF standards for valid values.
        payee : str, default=None
            The name of the payee on the other end of the transaction. The payee is the receiver of the money if amount is
            negative, else they are the sender of the money.
        payee_address : str, default=None
            The address of the aforementioned payee.
        category : Category, default=None
            The category object that represents the transaction.
        check_number : int, default=None
            The check number if the transaction relates to a check
        reimbursable_expense : bool, default=None
            Whether or not this transaction is flagged as a reimbursable business expense.
        small_business_expense : bool, default=None
            Whether or not this transaction is flagged as a small business expense.
        to_account : str, default=None
            The account the transaction was sent to, if applicable.
        first_payment_date : datetime.datetime, default=None
            If this transaction was completed over multiple days, the first payment date.
        loan_length : decimal.Decimal, default=None
            The length of the loan, if applicable.
        num_payments : int, default=None
            If this payment was split over multiple payments, the number of such payments.
        periods_per_annum : int, default=None
            The periods per annum for this transaction.
        interest_rate : decimal.Decimal, default=None
            The interest rate on this transaction.
        current_loan_balance : decimal.Decimal, default=None
            The current loan balance, if applicable.
        original_loan_amount : decimal.Decimal, default=None
            The original loan amount, if applicable.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF file.
        splits : TransactionList, default=None
            If this transaction has multiple categories (e.g. an Amazon purchase of an electrical item and a book), it
            can be split in QIF files to represent this. Each split has its own memo, category and amount.
            This parameter is a TransactionList containing the splits for this transaction.
        """
        self._date = date
        self._amount = amount
        self._memo = memo
        self._cleared = cleared
        self._payee = payee
        self._payee_address = payee_address
        self._category = category
        self._check_number = check_number
        self._reimbursable_expense = reimbursable_expense
        self._small_business_expense = small_business_expense
        self._to_account = to_account
        self._first_payment_date = first_payment_date
        self._loan_length = loan_length
        self._num_payments = num_payments
        self._periods_per_annum = periods_per_annum
        self._interest_rate = interest_rate
        self._current_loan_balance = current_loan_balance
        self._original_loan_amount = original_loan_amount
        self._line_number = line_number

        self._split_categories = {}

        if splits:
            self._assert_type(splits, Split)
            self._splits = TransactionList(*splits)
            self._is_split = True
            self._last_split = splits[-1]
            for split in splits:
                self._split_categories = create_categories(split.category, self._split_categories)
        else:
            self._splits = TransactionList()
            self._is_split = False
            self._last_split = None

    def __eq__(self, other):
        if not isinstance(other, Transaction):
            return False
        return self._date == other.date and self._amount == other.amount

    def __str__(self):
        properties = ''
        ignore = ['_last_split', '_is_split']
        for (object_property, value) in self.__dict__.items():
            if value and object_property not in ignore:
                if object_property == '_category':
                    properties += f'\n    Category: {value.name}'
                elif object_property == '_split_categories':
                    properties += f'\n    Split Categories: {[name for name in value]}'
                elif object_property == '_splits':
                    properties += f'\n    Splits: {len(value)} total split(s)'
                else:
                    properties += f'\n    {object_property.strip("_").replace("_", " ").title()}: {value}'

        return 'Transaction:' + properties

    def __repr__(self):
        properties = ''
        for (object_property, value) in self.__dict__.items():
            if value:
                properties += f'{object_property.strip("_")}={repr(value)}, '

        properties = properties.strip(', ')
        return f'Transaction({properties})'

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, new_date):
        if not isinstance(new_date, datetime):
            raise TypeError(f'New date must be datetime object')
        self._date = new_date

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, new_amount):
        try:
            self._amount = Decimal(new_amount)
        except ValueError:
            raise TypeError('Amount can only be int, float or decimal.Decimal')

    @property
    def memo(self):
        return self._memo

    @memo.setter
    def memo(self, new_memo):
        self._memo = str(new_memo)

    @property
    def cleared(self):
        return self._cleared

    @cleared.setter
    def cleared(self, new_str):
        self._cleared = str(new_str)

    @property
    def payee(self):
        return self._payee

    @payee.setter
    def payee(self, new_payee):
        self._payee = str(new_payee)

    @property
    def payee_address(self):
        return self._payee_address

    @payee_address.setter
    def payee_address(self, new_payee_address):
        self._payee_address = str(new_payee_address)

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, new_category):
        self._assert_type([new_category], Category)
        self._category = new_category

    @property
    def check_number(self):
        return self._check_number

    @check_number.setter
    def check_number(self, new_number):
        try:
            new_number = int(new_number)

            if new_number <= 0:
                raise ValueError
        except ValueError:
            raise TypeError('Check number must be a positive integer')

    @property
    def reimbursable_expense(self):
        return self._reimbursable_expense

    @reimbursable_expense.setter
    def reimbursable_expense(self, new_flag):
        if isinstance(new_flag, str) and new_flag.lower() == 'false':
            self._reimbursable_expense = False
        else:
            self._reimbursable_expense = True

    @property
    def small_business_expense(self):
        return self._small_business_expense

    @small_business_expense.setter
    def small_business_expense(self, new_flag):
        if isinstance(new_flag, str) and new_flag.lower() == 'false':
            self._small_business_expense = False
        else:
            self._small_business_expense = True

    @property
    def to_account(self):
        return self._to_account

    @to_account.setter
    def to_account(self, new_account):
        self._to_account = str(new_account)

    @property
    def first_payment_date(self):
        return self._first_payment_date

    @first_payment_date.setter
    def first_payment_date(self, new_date):
        self._first_payment_date = parse_date(new_date)

    @property
    def loan_length(self):
        return self._loan_length

    @loan_length.setter
    def loan_length(self, new_length):
        self._loan_length = Decimal(new_length)

    @property
    def num_payments(self):
        return self._num_payments

    @num_payments.setter
    def num_payments(self, new_num):
        self._num_payments = int(new_num)

    @property
    def periods_per_annum(self):
        return self._periods_per_annum

    @periods_per_annum.setter
    def periods_per_annum(self, new_num):
        self._periods_per_annum = int(new_num)

    @property
    def interest_rate(self):
        return self._interest_rate

    @interest_rate.setter
    def interest_rate(self, new_rate):
        self._interest_rate = Decimal(new_rate)

    @property
    def current_loan_balance(self):
        return self._current_loan_balance

    @current_loan_balance.setter
    def current_loan_balance(self, new_balance):
        self._current_loan_balance = Decimal(new_balance)

    @property
    def original_loan_amount(self):
        return self._original_loan_amount

    @original_loan_amount.setter
    def original_loan_amount(self, new_amount):
        self._original_loan_amount = Decimal(new_amount)

    @property
    def splits(self):
        return self._splits

    @splits.setter
    def splits(self, new_splits):
        self._assert_type(new_splits, Split)
        self._splits = new_splits
        self.refresh_is_split()

    @property
    def is_split(self):
        return self._is_split

    def refresh_is_split(self):
        self._is_split = bool(self._splits)

    @property
    def split_categories(self):
        return self._split_categories

    @property
    def line_number(self):
        return self._line_number

    @classmethod
    def from_list(cls, lst, day_first=True, line_number=None):
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the transaction.
        day_first : bool, default=True
             Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF file.

        Returns
        -------
        Transaction
            A Transaction object created from the QIF strings.
        """
        kwargs = {}
        categories = {}
        classes = []
        splits = TransactionList()
        current_split = None
        split_categories = {}
        for field in lst:
            field = field.replace('\n', '')

            if not field:
                continue
            line_code = field[0]

            try:
                field_info = field[1:]
            except KeyError:
                field_info = ''

            # Check the QIF line code for banking-related operations, then append to kwargs.
            if line_code == 'S':
                new_split = Split()
                category_name = field_info.split(':')[-1]
                new_category = Category(category_name)
                new_category.hierarchy = field_info
                split_categories = create_categories(new_category, split_categories)
                new_split.category = new_category
                splits.append(new_split)
                current_split = new_split
            elif line_code == 'D':
                transaction_date = parse_date(field_info, day_first)
                if not splits:
                    kwargs['date'] = transaction_date
                else:
                    current_split.date = transaction_date
            elif line_code == 'E':
                current_split.memo = field_info
            elif line_code == '$' or line_code == '£':
                current_split.amount = Decimal(field_info.replace(',', ''))
            elif line_code == '%':
                current_split.percent = Decimal(field_info.split(' ')[0].replace('%', ''))
            elif line_code == 'T' or line_code == 'U':
                if not splits:
                    kwargs['amount'] = Decimal(field_info.replace(',', ''))
                else:
                    current_split.amount = Decimal(field_info.replace(',', ''))
            elif line_code == 'M':
                if not splits:
                    kwargs['memo'] = field_info
                else:
                    current_split.memo = field_info
            elif line_code == 'C':
                if not splits:
                    kwargs['cleared'] = field_info
                else:
                    current_split.cleared = field_info
            elif line_code == 'P':
                kwargs['payee'] = field_info
            elif line_code == 'A':
                if not splits:
                    kwargs['payee_address'] = field_info
                else:
                    current_split.payee_address = field_info
            elif line_code == 'L':
                # 'L' can store classes, denoted by '/'
                if '/' in field_info:
                    field_info, class_name = field_info.split('/')[0], field_info.split('/')[1]
                    classes.append(Class(name=class_name))

                # 'L' can represent both categories and the 'to' transfer account
                # Transfer accounts are denoted by [ ] so check for those
                if field_info.startswith('['):
                    if not splits:
                        kwargs['to_account'] = field_info[1:-1]
                    else:
                        current_split.to_account = field_info[1:-1]
                else:
                    category_name = field_info.split(':')[-1]
                    new_category = Category(category_name)
                    new_category.hierarchy = field_info
                    categories = create_categories(new_category, categories)
                    if not splits:
                        kwargs['category'] = new_category
                    else:
                        current_split.category = new_category
            elif line_code == 'N':
                if not splits:
                    kwargs['check_number'] = int(field_info)
                else:
                    current_split.check_number = int(field_info)
            elif line_code == 'F':
                if field_info.lower() == 'false':
                    kwargs['reimbursable_expense'] = False
                else:
                    kwargs['reimbursable_expense'] = True
            elif line_code == '1':
                kwargs['first_payment_date'] = parse_date(field_info, day_first)
            elif line_code == '2':
                kwargs['loan_length'] = Decimal(field_info.replace(',', ''))
            elif line_code == '3':
                kwargs['num_payments'] = int(field_info.replace(',', ''))
            elif line_code == '4':
                kwargs['periods_per_annum'] = int(field_info.replace(',', ''))
            elif line_code == '5':
                kwargs['interest_rate'] = Decimal(field_info.replace(',', ''))
            elif line_code == '6':
                kwargs['current_loan_balance'] = Decimal(field_info.replace(',', ''))
            elif line_code == '7':
                kwargs['original_loan_amount'] = Decimal(field_info.replace(',', ''))

        if line_number is not None:
            kwargs['line_number'] = line_number

        # Set splits percentage if they don't already have one
        if splits:
            total_amount = kwargs['amount']
            for split in splits:
                if not split.percent:
                    split.percent = abs(split.amount / total_amount) * 100

        categories.update({name: category for (name, category) in split_categories.items() if name not in categories})
        kwargs['splits'] = splits
        return cls(**kwargs), categories, classes

    @classmethod
    def from_string(cls, string, separator='\n', day_first=True, line_number=None):
        """Return a class instance from a QIF file section string.

        Parameters
        ----------
        string : str
            The string containing the QIF-formatted data.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.
        day_first : bool, default=True
             Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the transaction in the QIF file.

        Returns
        -------
        Transaction
            A Transaction object created from the QIF strings.
        """
        property_list = string.split(separator)
        return cls.from_list(property_list, day_first, line_number)

    @staticmethod
    def _assert_type(iterable, types):
        # Assert that all items in a given list are of specific types.
        if isinstance(iterable, dict):
            for item in iterable.values():
                if not isinstance(item, types):
                    raise TypeError(f'\'{repr(item)} is not of type {types}')
        else:
            for item in iterable:
                if not isinstance(item, types):
                    raise TypeError(f'\'{repr(item)} is not of type {types}')

    def add_split(self, new_split):
        """Add a Split to Transaction.

        Parameters
        ----------
        new_split : Split
            The Split object to be added.

        Raises
        -------
        TypeError
            If ``new_split`` is not a Split.
        ValueError
            If percentage sum of all splits added is greater than 100.
        ValueError
            If sum of split amounts is greater than overall transaction amount.
        """
        if not isinstance(new_split, Split):
            raise TypeError('Must be a Split object')

        if new_split.percent and sum([split.percent for split in self._splits
                                      if split.percent]) + new_split.percent > 100:
            raise ValueError('Percentage sum of splits cannot add up to more than 100%')

        if new_split.amount and sum([split.amount for split in self._splits if split.amount]) > self._amount:
            raise ValueError('Splits amounts cannot sum to more than Transaction amount')

        self._splits.append(new_split)
        self.refresh_is_split()

    def remove_split(self, multiple=False, **kwargs):
        """Remove Split(s) from Transaction if they match a set of kwargs.

        Parameters
        ----------
        multiple : bool, default=None
            Whether or not to remove multiple splits that match ``kwargs``.
        kwargs
            The keyword arguments to filter the splits by (e.g. amount=150 would only remove splits with amount=150)

        Raises
        -------
        RuntimeError
            If no kwargs are provided.
        """
        if not kwargs:
            raise RuntimeError('No kwargs provided')

        indices = []

        for (i, split) in enumerate(self._splits):
            found_split = False
            for (key, value) in kwargs.items():
                if not key.startswith('_'):
                    key = '_' + key
                if split.__dict__.get(key) != value:
                    found_split = False
                    break
                found_split = True

            if found_split:
                indices.append(i)
                if not multiple:
                    break

        # Reverse so indices in list don't change as items get removed
        for idx in reversed(indices):
            self._splits.pop(idx)
        self.refresh_is_split()

    def to_dict(self, ignore=None, dictify_splits=True, dictify_category=True):
        """Return a dict object representing the Transaction.

        Parameters
        ----------
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.
        dictify_splits : bool, default=True
             Whether Splits within Transaction objects should be converted to dicts or left as they are.
        dictify_category : bool, default=True
             Whether the Transaction's category should be converted to a dict or left as it is.

        Returns
        -------
        dict
            A dict representing the Transaction object.
        """
        if ignore is None:
            ignore = []

        res = {key.strip('_'): value for (key, value) in self.__dict__.items()
               if value and key.strip('_') not in ignore}

        if dictify_splits and 'splits' not in ignore and self._is_split:
            res['splits'] = [split.to_dict(ignore=ignore) for split in res['splits']]

        if dictify_category and 'category' not in ignore and self._category is not None:
            res['category'] = self._category.to_dict()

        return res


class Split:
    """
    A class used to represent a split in a transaction.

    Examples
    --------
    Adding Splits to a transaction to show that there were two categories that represent the transaction.

    >>> import quiffen
    >>> from datetime import datetime
    >>> tr = quiffen.Transaction(date=datetime.now(), amount=150.60)
    >>> beauty = quiffen.Category('Beauty')
    >>> electrical = quiffen.Category('Electrical')
    >>> splt1 = quiffen.Split(amount=50, category=beauty)
    >>> splt2 = quiffen.Split(amount=100.60, category=electrical)
    >>> tr.add_split(splt1)
    >>> tr.add_split(splt2)
    >>> print(tr)
    Transaction:
        Date: 2021-07-05 10:59:02.456190
        Amount: 150.6
        Splits: 2 total split(s)
    >>> print(tr.splits)
    [Split(amount=50, category=Category(name='Beauty', expense=True, hierarchy='Beauty')), Split(amount=100.6,
    category=Category(name='Electrical', expense=True, hierarchy='Electrical'))]
    >>> tr.remove_split(amount=50)
    >>> print(tr.splits)
    [Split(amount=100.6, category=Category(name='Electrical', expense=True, hierarchy='Electrical'))]
    """

    def __init__(self,
                 date: datetime = None,
                 amount: Decimal = None,
                 memo: str = None,
                 cleared: str = None,
                 payee_address: str = None,
                 category: Category = None,
                 to_account: str = None,
                 check_number: int = None,
                 percent: Decimal = None
                 ):
        """Initialise an instance of the Split class.

        Parameters
        ----------
        date : datetime, default=None
            The date of the split transaction.
        amount : decimal.Decimal, default=None
            The amount this of this split.
        memo : str, default=None
            The memo (reference) for this split.
        cleared : str, default=None
            The cleared status of this split. See the QIF standards for valid values.
        payee_address : str, default=None
            The address of the payee for this split.
        category : Category, default=None
            The Category object that represents the category this split falls under.
        to_account : str, default=None
            The to account for this split.
        check_number : int, default=None
            The check number if this transaction relates to a check.
        percent : decimal.Decimal, default=None
            The percentage value of this split compared to the overall transaction.
        """
        self._date = date
        self._amount = amount
        self._memo = memo
        self._cleared = cleared
        self._payee_address = payee_address
        self._category = category
        self._to_account = to_account
        self._check_number = check_number
        self._percent = percent

    def __eq__(self, other):
        if not isinstance(other, Split):
            return False
        return self.__dict__ == other.__dict__

    def __str__(self):
        properties = ''
        for (object_property, value) in self.__dict__.items():
            if value:
                if object_property == '_category':
                    properties += f'\n        Category: {value.name}'
                else:
                    properties += f'\n        {object_property.strip("_").replace("_", " ").title()}: {value}'

        return '\n    Split:' + properties

    def __repr__(self):
        properties = ''
        for (object_property, value) in self.__dict__.items():
            if value:
                properties += f'{object_property.strip("_")}={repr(value)}, '

        properties = properties.strip(', ')
        return f'Split({properties})'

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, new_date):
        if not isinstance(new_date, datetime):
            raise TypeError(f'New date must be datetime object')
        self._date = new_date

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, new_amount):
        try:
            self._amount = Decimal(new_amount)
        except ValueError:
            raise TypeError('Amount can only be int, float or Decimal')

    @property
    def memo(self):
        return self._memo

    @memo.setter
    def memo(self, new_memo):
        self._memo = str(new_memo)

    @property
    def cleared(self):
        return self._cleared

    @cleared.setter
    def cleared(self, new_val):
        self._cleared = str(new_val)

    @property
    def payee_address(self):
        return self._payee_address

    @payee_address.setter
    def payee_address(self, new_payee_address):
        self._payee_address = str(new_payee_address)

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, new_category):
        if not isinstance(new_category, Category):
            raise TypeError('Must be a Category object')
        self._category = new_category

    @property
    def to_account(self):
        return self._to_account

    @to_account.setter
    def to_account(self, new_account):
        self._to_account = str(new_account)

    @property
    def check_number(self):
        return self._check_number

    @check_number.setter
    def check_number(self, new_number):
        try:
            new_number = int(new_number)

            if new_number <= 0:
                raise ValueError
        except ValueError:
            raise TypeError('Check number must be a positive integer')

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, new_percent):
        self._percent = Decimal(new_percent)

    def to_dict(self, ignore=None, dictify_category=True):
        """Return a dict object representing the Split.

        Parameters
        ----------
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.
        dictify_category : bool, default=True
             Whether the Split's category should be converted to a dict or left as it is.

        Returns
        -------
        dict
            A dict representing the Split object.
        """
        if ignore is None:
            ignore = []
        res = {key.strip('_'): value for (key, value) in self.__dict__.items()
               if value is not None and key.strip('_') not in ignore}

        if dictify_category and 'category' not in ignore and self._category is not None:
            res['category'] = self._category.to_dict()

        return res


class Investment:
    """
    A class used to represent an investment.

    Acts almost identically to the Transaction class, but with a few different parameters and no splits.
    """

    def __init__(self,
                 date: datetime,
                 action: str = None,
                 security: str = None,
                 price: Decimal = None,
                 quantity: Decimal = None,
                 cleared: str = None,
                 amount: Decimal = None,
                 memo: str = None,
                 first_line: str = None,
                 to_account: str = None,
                 transfer_amount: float = None,
                 commission: float = None,
                 line_number: int = None
                 ):
        """Initialise an instance of the Investment class.

        Parameters
        ----------
        date : datetime
            Date investment occurred. May or may not include timestamp.
        action : str, default=None
            The investment action (Buy, Sell, etc.)
        security : str, default=None
            The security name.
        price : decimal.Decimal, default=None
            The price of the security.
        quantity : decimal.Decimal, default=None
            The quantity of the security bought, sold, etc.
        cleared : str, default=None
            The cleared status of this investment. See the QIF standards for valid values.
        amount : decimal.Decimal, default=None
            The overall amount of this investment.
        memo : str, default=None
            Also known as the reference. A string describing the purpose behind the investment.
        first_line : str, default=None
            The first line of the investment.
        to_account : str, default=None
            The to account of the investment, if applicable.
        transfer_amount : decimal.Decimal, default=None
            The amount transferred for the investment.
        commission : decimal.Decimal, default=None
            The commission paid/received on the investment.
        line_number : int, default=None
            The line number of the investment in the QIF file.
        """
        self._date = date
        self._action = action
        self._security = security
        self._price = price
        self._quantity = quantity
        self._cleared = cleared
        self._amount = amount
        self._memo = memo
        self._first_line = first_line
        self._to_account = to_account
        self._transfer_amount = transfer_amount
        self._commission = commission
        self._line_number = line_number

    def __eq__(self, other):
        if not isinstance(other, Investment):
            return False

        return self.__dict__ == other.__dict__

    def __str__(self):
        properties = ''
        for (object_property, value) in self.__dict__.items():
            if value:
                properties += f'\n    {object_property.strip("_").replace("_", " ").title()}: {value}'

        return 'Investment:' + properties

    def __repr__(self):
        properties = ''
        for (object_property, value) in self.__dict__.items():
            if value is not None:
                properties += f'{object_property.strip("_")}={repr(value)}, '

        properties = properties.strip(', ')
        return f'Investment({properties})'

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, new_date):
        if not isinstance(new_date, datetime):
            raise TypeError(f'New date must be datetime object')
        self._date = new_date

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, new_action):
        self._action = str(new_action)

    @property
    def security(self):
        return self._security

    @security.setter
    def security(self, new_security):
        self._security = str(new_security)

    @property
    def price(self):
        return self._price

    @price.setter
    def price(self, new_price):
        try:
            self._price = Decimal(new_price)
        except ValueError:
            raise TypeError('Price can only be int, float or decimal.Decimal')

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, new_quantity):
        try:
            self._quantity = Decimal(new_quantity)
        except ValueError:
            raise TypeError('Quantity can only be int, float or decimal.Decimal')

    @property
    def cleared(self):
        return self._cleared

    @cleared.setter
    def cleared(self, new_value):
        self._cleared = str(new_value)

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, new_amount):
        try:
            self._amount = Decimal(new_amount)
        except ValueError:
            raise TypeError('Amount can only be int, float or decimal.Decimal')

    @property
    def memo(self):
        return self._memo

    @memo.setter
    def memo(self, new_memo):
        self._memo = str(new_memo)

    @property
    def first_line(self):
        return self._first_line

    @first_line.setter
    def first_line(self, new_first_line):
        self._first_line = str(new_first_line)

    @property
    def to_account(self):
        return self._to_account

    @to_account.setter
    def to_account(self, new_to_account):
        self._to_account = str(new_to_account)

    @property
    def transfer_amount(self):
        return self._transfer_amount

    @transfer_amount.setter
    def transfer_amount(self, new_transfer_amount):
        try:
            self._transfer_amount = Decimal(new_transfer_amount)
        except ValueError:
            raise TypeError('Transfer amount can only be int, float or decimal.Decimal')

    @property
    def commission(self):
        return self._commission

    @commission.setter
    def commission(self, new_commission):
        try:
            self._commission = Decimal(new_commission)
        except ValueError:
            raise TypeError('Commission can only be int, float or decimal.Decimal')

    @property
    def line_number(self):
        return self._line_number

    @classmethod
    def from_list(cls, lst, day_first=True, line_number=None):
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the investment.
        day_first : bool, default=True
            Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the investment in the QIF file.

        Returns
        -------
        Investment
            An Investment object created from the QIF strings.
        """
        kwargs = {}
        for field in lst:
            field = field.replace('\n', '')

            if not field:
                continue
            line_code = field[0]

            try:
                field_info = field[1:]
            except KeyError:
                field_info = ''

            # Check the QIF line code for banking-related operations, then append to kwargs.
            if line_code == 'D':
                transaction_date = parse_date(field_info, day_first)
                kwargs['date'] = transaction_date
            elif line_code == 'N':
                kwargs['action'] = field_info
            elif line_code == 'Y':
                kwargs['security'] = field_info
            elif line_code == 'I':
                kwargs['price'] = Decimal(field_info.replace(',', ''))
            elif line_code == 'Q':
                kwargs['quantity'] = Decimal(field_info.replace(',', ''))
            elif line_code == 'C':
                kwargs['cleared'] = field_info
            elif line_code == 'T' or line_code == 'U':
                kwargs['amount'] = Decimal(field_info.replace(',', ''))
            elif line_code == 'M':
                kwargs['memo'] = field_info
            elif line_code == 'P':
                kwargs['first_line'] = field_info
            elif line_code == 'L':
                kwargs['to_account'] = field_info
            elif line_code == '$':
                kwargs['transfer_amount'] = Decimal(field_info.replace(',', ''))
            elif line_code == 'O':
                kwargs['commission'] = Decimal(field_info.replace(',', ''))

        if line_number is not None:
            kwargs['line_number'] = line_number

        return cls(**kwargs)

    @classmethod
    def from_string(cls, string, separator='\n', day_first=True, line_number=None):
        """Return a class instance from a QIF file section string.

        Parameters
        ----------
        string : str
            The string containing the QIF-formatted data.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.
        day_first : bool, default=True
             Whether the day or month comes first in the date.
        line_number : int, default=None
            The line number of the header line of the investment in the QIF file.

        Returns
        -------
        Investment
            An Investment object created from the QIF strings.
        """
        property_list = string.split(separator)
        return cls.from_list(property_list, day_first, line_number)

    def to_dict(self, ignore=None):
        """Return a dict object representing the Transaction.

        Parameters
        ----------
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.

        Returns
        -------
        dict
            A dict representing the Investment object.
        """
        if ignore is None:
            ignore = []
        return {key.strip('_'): value for (key, value) in self.__dict__.items()
                if value is not None and key.strip('_') not in ignore}



from decimal import Decimal


class Category:
    """
    A node-like class used to represent a category. Can be built into trees to represent category families.

    Attributes
    ----------
    children : list of Category
        A list of the category's children objects.

    Examples
    --------
    Creating a category tree, then rendering it to console.

    >>> import quiffen
    >>> food = quiffen.Category('Food')
    >>> food
    Category(name='Food', expense=True, hierarchy='Food')
    >>> essentials = quiffen.Category('Essentials')
    >>> food.add_child(essentials)
    >>> pastas = quiffen.Category('Pastas')
    >>> essentials.add_child(pastas)
    >>> pastas.hierarchy
    'Food:Essentials:Pastas'
    >>> meat = quiffen.Category('Meat')
    >>> food.add_child(meat)
    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Meat

    Removing a category from a tree by both passing in the Category instance or the category name string.

    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Meat
       └─ Chicken
    >>> meat
    Category(name='Meat', expense=True, parent=Category(name='Food', expense=True, hierarchy='Food'), hierarchy='Food:Meat')
    >>> food.remove_child(meat, keep_children=True)
    >>> print(food.render_tree())
    Food (root)
    └─ Essentials
       └─ Pastas
    └─ Chicken
    >>> food.remove_child('Essentials')
    >>> print(food.render_tree())
    Food (root)
    └─ Chicken
    """

    def __init__(self,
                 name: str,
                 desc: str = None,
                 tax_related: bool = None,
                 expense: bool = True,
                 income: bool = False,
                 budget_amount: Decimal = None,
                 tax_schedule_info: str = None,
                 parent=None,
                 hierarchy: str = None):
        """Initialise an instance of the Category class.

        Parameters
        ----------
        name : str
            The name of the category.
        desc : str, default=None
            The category's description.
        tax_related : bool, default=None
            Whether the category represents a tax related group of transactions.
        expense : bool, default=True
            Whether the category represents expenses as opposed to income.
        income : bool, default=False
            Whether the category represents income as opposed to expenses.
        budget_amount : decimal.Decimal, default=None
            The budget amount for this category.
        tax_schedule_info : str, default=None
            Information about the tax schedule for this category.
        parent : Category, default=None
            The parent category for this category.
        hierarchy : str, default=None
            The category hierarchy, separated by : characters.
        """
        self._name = name
        self._desc = desc
        self._tax_related = tax_related
        self._expense = expense
        self._income = income
        self._budget_amount = budget_amount
        self._tax_schedule_info = tax_schedule_info
        self._parent = parent

        if hierarchy is not None:
            self._hierarchy = hierarchy
        else:
            self._hierarchy = name

        self._children = []

    def __eq__(self, other):
        if not isinstance(other, Category):
            return False
        return self._name == other.name

    def __str__(self):
        return repr(self)

    def __repr__(self):
        properties = ''
        ignore = ['_children']
        for (object_property, value) in self.__dict__.items():
            if value and object_property not in ignore:
                properties += f'{object_property.strip("_")}={repr(value)}, '

        properties = properties.strip(', ')
        return f'Category({properties})'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        self._name = str(new_name)

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, new_desc):
        self._desc = str(new_desc)

    @property
    def tax_related(self):
        return self._tax_related

    @tax_related.setter
    def tax_related(self, new_bool):
        if isinstance(new_bool, str) and new_bool.lower() == 'false':
            self._tax_related = False
        else:
            self._tax_related = bool(new_bool)

    @property
    def expense(self):
        return self._expense

    @expense.setter
    def expense(self, new_bool):
        if isinstance(new_bool, str) and new_bool.lower() == 'false':
            self._expense = False
            self._income = True
        else:
            self._expense = bool(new_bool)
            self._income = not bool(new_bool)

    @property
    def income(self):
        return self._income

    @income.setter
    def income(self, new_bool):
        if isinstance(new_bool, str) and new_bool.lower() == 'false':
            self._income = False
            self._expense = True
        else:
            self._income = bool(new_bool)
            self._expense = not bool(new_bool)

    @property
    def budget_amount(self):
        return self._budget_amount

    @budget_amount.setter
    def budget_amount(self, new_amount):
        self._budget_amount = Decimal(new_amount)

    @property
    def tax_schedule_info(self):
        return self._tax_schedule_info

    @tax_schedule_info.setter
    def tax_schedule_info(self, new_info):
        self._tax_schedule_info = str(new_info)

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        if not isinstance(new_parent, (Category, type(None))):
            raise TypeError('New parent must be Category object')
        self._parent = new_parent
        if self._parent and self not in self._parent.children:
            new_parent.add_child(self)

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, new_children):
        # Check types
        for child in new_children:
            if not isinstance(child, (Category, type(None))):
                raise TypeError('Children must be Category objects')

            if child.hierarchy is None:
                child.hierarchy = child.name

            if self._hierarchy not in child.hierarchy:
                child.hierarchy = f'{self._hierarchy}:{child.hierarchy}'

        self._children = new_children

        # Update parent nodes
        for child in self._children:
            child.parent = self

    @property
    def hierarchy(self):
        return self._hierarchy

    @hierarchy.setter
    def hierarchy(self, new_hierarchy):
        if new_hierarchy.split(':')[-1] != self._name:
            raise RuntimeError('Invalid hierarchy. Must end with current category.')
        self._hierarchy = new_hierarchy

    @classmethod
    def from_list(cls, lst):
        """Return a Category instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the category.

        Returns
        -------
        Category
            A Category object created from the QIF strings.
        """
        kwargs = {}
        for field in lst:
            field = field.replace('\n', '')

            if not field:
                continue
            line_code = field[0]

            try:
                field_info = field[1:]
            except KeyError:
                field_info = ''

            if line_code == 'N':
                categories = field_info.split(':')
                kwargs['name'] = categories[-1]
                kwargs['hierarchy'] = field_info
            elif line_code == 'D':
                kwargs['desc'] = field_info
            elif line_code == 'T':
                if field_info.lower() == 'false':
                    kwargs['tax_related'] = False
                else:
                    kwargs['tax_related'] = True
            elif line_code == 'E':
                kwargs['expense'] = True
                kwargs['income'] = False
            elif line_code == 'I':
                kwargs['expense'] = False
                kwargs['income'] = True
            elif line_code == 'B':
                kwargs['budget_amount'] = Decimal(field_info.replace(',', ''))
            elif line_code == 'R':
                kwargs['tax_schedule_info'] = field_info

        return cls(**kwargs)

    @classmethod
    def from_string(cls, string, separator='\n'):
        """Return a Category instance from a QIF file section string

        Parameters
        ----------
        string : str
            The string containing the QIF-formatted data.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.

        Returns
        -------
        Category
            A Category object created from the QIF strings.
        """
        property_list = string.split(separator)
        return cls.from_list(property_list)

    def add_child(self, child_category):
        """Add a child category to current category.

        Parameters
        ----------
        child_category : str or Category
            A Category object to be added as a child or the string name of a new category to be added.

        Raises
        ------
        TypeError
            If ``child_category`` is not a str or Category.
        """
        if isinstance(child_category, str):
            child_category = Category(child_category)

        if not isinstance(child_category, Category):
            raise TypeError('Child must be of type Category or a string')

        self._children.append(child_category)
        child_category = self._children[-1]

        if child_category.parent != self:
            child_category.parent = self

        if child_category.hierarchy is None:
            child_category.hierarchy = child_category.name

        if self._hierarchy not in child_category.hierarchy:
            child_category.hierarchy = f'{self._hierarchy}:{child_category.hierarchy}'

    def remove_child(self, child_category, keep_children=False):
        """Remove a child category from the current category's hierarchy.

        Parameters
        ----------
        child_category : str or Category
            The Category object to be removed as a child or the string name of the category.
        keep_children : bool, default=False
            Whether or not the children of the removed category should be kept in the tree and moved up a level.

        Raises
        ------
        TypeError
            If ``child_category`` is not a str or Category.
        """
        if isinstance(child_category, str):
            name = child_category
        elif isinstance(child_category, Category):
            name = child_category.name
        else:
            raise TypeError('Child must be of type Category or a string')

        child_category = self.find_category(child_category)
        parent = child_category.parent
        child_category.parent = None
        child_category.hierarchy.replace(self._hierarchy + ':', '')

        if keep_children:
            for child in child_category.children:
                child.hierarchy = child.hierarchy.replace(child_category.name + ':', '')
                parent.add_child(child)

        parent.children = [child for child in parent.children if child is not child_category]

    def traverse_down(self):
        """Return a flat list of all children, grandchildren etc.

        Returns
        -------
        all_children : list of Category
        """
        nodes_to_visit = [self]
        all_children = []
        while nodes_to_visit:
            current_node = nodes_to_visit.pop()
            all_children.append(current_node)
            nodes_to_visit.extend(current_node.children)
        return all_children

    def traverse_up(self):
        """Return a list of all parents, grandparents etc.

        Returns
        -------
        all_parents : list of Category
        """
        current_node = self
        all_parents = []
        while current_node._parent:
            all_parents.append(current_node._parent)
            current_node = current_node._parent
        return all_parents

    def render_tree(self, _level=0):
        """Renders a tree-like structure for categories.

        Runs recursively, and uses ``_level`` just to keep track of the indentation.
        """
        if not self._children:
            return self._name

        if self._parent is None:
            is_root_str = ' (root)'
        else:
            is_root_str = ''

        return self._name + f'{is_root_str}\n' + '\n'.join(['   ' * _level + '└─ ' + child.render_tree(_level + 1)
                                                            for child in self._children])

    def find_category(self, node_name):
        """Returns a node with a given name from the children of current node.

        Parameters
        ----------
        node_name : str
            The name of the category to find.

        Returns
        -------
        Category
            The category searched for.

        Raises
        ------
        KeyError
            If the category cannot be found.
        """
        nodes_to_visit = [self]
        while nodes_to_visit:
            current_node = nodes_to_visit.pop()
            if current_node.name == node_name:
                return current_node
            nodes_to_visit.extend(current_node.children)
        raise KeyError(f'Node with name \'{node_name}\' not found')

    def to_dict(self, ignore=None):
        """Return a representation of the Category object as a dict.

        Parameters
        ----------
        ignore : list of str
            A list of the object's attributes that should not be included in the resulting dict.

        Returns
        -------
        dict
        """
        if ignore is None:
            ignore = []

        res = {key.strip('_'): value for (key, value) in self.__dict__.items()
               if key.strip('_') not in ignore and value is not None}

        if self._children and 'children' not in ignore:
            res['children'] = [category.name for category in self._children]

        if self._parent and 'parent' not in ignore:
            res['parent'] = self._parent.name

        return res


class Class:
    """
    A class used to represent a QIF Class.
    """

    def __init__(self,
                 name: str,
                 desc: str = None):
        """Initialise an instance of the Class class.

        Parameters
        ----------
        name : str
            The name of the class.
        desc : str, default=None
            The description of the class.
        """
        self._name = name
        self._desc = desc

    def __eq__(self, other):
        return self._name == other.name

    def __str__(self):
        res = f'Class:\n    Name: {self._name}'
        if self._desc:
            res += f'\n    Description: {self._desc}'
        return res

    def __repr__(self):
        if self._desc:
            return f'Class(name={repr(self._name)}, desc={repr(self._desc)})'
        else:
            return f'Class(name={repr(self._name)})'

    @property
    def name(self):
        return self._name

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, new_desc):
        self._desc = str(new_desc)

    @classmethod
    def from_list(cls, lst):
        """Return a class instance from a list of QIF strings.

        Parameters
        ----------
        lst : list of str
            List of strings containing QIF information about the account.

        Returns
        -------
        Class
            A Class object created from the QIF strings.
        """
        name = None
        desc = None

        for field in lst:
            field = field.replace('\n', '')

            if not field:
                continue
            line_code = field[0]

            try:
                field_info = field[1:]
            except KeyError:
                field_info = ''

            if line_code == 'N':
                name = field_info
            elif line_code == 'D':
                desc = field_info

        if not name:
            raise RuntimeError('No name specified for Class')

        return cls(name, desc)

    @classmethod
    def from_string(cls, string, separator='\n'):
        """Return a class instance from a QIF file section string.

        Parameters
        ----------
        string : str
            The string containing the QIF-formatted data.
        separator : str, default='\n'
             The line separator for the QIF file. This probably won't need changing.

        Returns
        -------
        Class
            A Class object created from the QIF strings.
        """
        property_list = string.split(separator)
        return cls.from_list(property_list)

    def to_dict(self, ignore=None):
        """Return a dict object representing the Class.

        Parameters
        ----------
        ignore : list of str, default=None
             A list of strings of parameters that should be excluded from the dict.

        Returns
        -------
        dict
            A dict representing the Class object.
        """
        if ignore is None:
            ignore = []

        return {key.strip('_'): value for (key, value) in self.__dict__.items()
                if key.strip('_') not in ignore and value is not None}
